// SPDX-License-Identifier: GPL-2.0-or-later
/* Copyright 2011 (c) Greg Kroah-Hartman <gregkh@suse.de> */
#ifndef _LSUSB_H
#define _LSUSB_H

extern int lsusb_t(void);
extern unsigned int verblevel;

#endif
